@@ -1,6 +1,7 @@
 from io import BytesIO
 import os
 import shutil
+from typing import Optional, Union
 from urllib.request import urlopen
 from zipfile import ZipFile
 
@@ -27,6 +28,21 @@ def echo_header(title):
     click.echo(click.style("-" * len(title), fg="green"))
 
 
+def cast_int_like_to_int(cli_arg: Union[None, str, int]) -> Union[None, str, int]:
+    """Cast int-like objects to int if possible
+
+    If the arg cannot be cast to an integer, return the unmodified object instead."""
+    try:
+        cli_arg_int = int(cli_arg)
+        return cli_arg_int
+    except TypeError:
+        # Don't cast if None
+        return cli_arg
+    except ValueError:
+        # Don't cast non-int-like strings
+        return cli_arg
+
+
 @click.group()
 def fab():
     """ FAB flask group commands"""
@@ -146,6 +162,29 @@ def create_db():
     click.echo(click.style("DB objects created", fg="green"))
 
 
+@fab.command("export-roles")
+@with_appcontext
+@click.option("--path", "-path", help="Specify filepath to export roles to")
+@click.option("--indent", help="Specify indent of generated JSON file")
+def export_roles(
+    path: Optional[str] = None, indent: Optional[Union[int, str]] = None
+) -> None:
+    """Exports roles with permissions and view menus to JSON file"""
+    # Cast negative numbers to int (as they are passed as str from CLI)
+    cast_indent = cast_int_like_to_int(indent)
+    current_app.appbuilder.sm.export_roles(path=path, indent=cast_indent)
+
+
+@fab.command("import-roles")
+@with_appcontext
+@click.option(
+    "--path", "-p", help="Path to a JSON file containing roles", required=True
+)
+def import_roles(path: str) -> None:
+    """ Imports roles with permissions and view menus from JSON file """
+    current_app.appbuilder.sm.import_roles(path)
+
+
 @fab.command("version")
 @with_appcontext
 def version():
