@@ -1,13 +1,15 @@
 import json
 import logging
-from typing import Set
+from typing import Any, Dict, List, Optional, Set
 import unittest
 
-from flask import Flask
+from flask import Flask, Response
+from flask.testing import FlaskClient
 from flask_appbuilder import AppBuilder, SQLA
 from flask_appbuilder.const import (
     API_SECURITY_PASSWORD_KEY,
     API_SECURITY_PROVIDER_KEY,
+    API_SECURITY_REFRESH_KEY,
     API_SECURITY_USERNAME_KEY,
     API_SECURITY_VERSION,
 )
@@ -17,26 +19,22 @@ import jinja2
 class FABTestCase(unittest.TestCase):
     @staticmethod
     def auth_client_get(client, token, uri):
-        return client.get(uri, headers={"Authorization": "Bearer {}".format(token)})
+        return client.get(uri, headers={"Authorization": f"Bearer {token}"})
 
     @staticmethod
     def auth_client_delete(client, token, uri):
-        return client.delete(uri, headers={"Authorization": "Bearer {}".format(token)})
+        return client.delete(uri, headers={"Authorization": f"Bearer {token}"})
 
     @staticmethod
     def auth_client_put(client, token, uri, json):
-        return client.put(
-            uri, json=json, headers={"Authorization": "Bearer {}".format(token)}
-        )
+        return client.put(uri, json=json, headers={"Authorization": f"Bearer {token}"})
 
     @staticmethod
     def auth_client_post(client, token, uri, json):
-        return client.post(
-            uri, json=json, headers={"Authorization": "Bearer {}".format(token)}
-        )
+        return client.post(uri, json=json, headers={"Authorization": f"Bearer {token}"})
 
     @staticmethod
-    def _login(client, username, password):
+    def _login(client, username, password, refresh: bool = False):
         """
             Login help method
         :param client: Flask test client
@@ -45,33 +43,53 @@ class FABTestCase(unittest.TestCase):
         :return: Flask client response class
         """
         return client.post(
-            "api/{}/security/login".format(API_SECURITY_VERSION),
-            data=json.dumps(
-                {
-                    API_SECURITY_USERNAME_KEY: username,
-                    API_SECURITY_PASSWORD_KEY: password,
-                    API_SECURITY_PROVIDER_KEY: "db",
-                }
-            ),
-            content_type="application/json",
+            f"api/{API_SECURITY_VERSION}/security/login",
+            json={
+                API_SECURITY_USERNAME_KEY: username,
+                API_SECURITY_PASSWORD_KEY: password,
+                API_SECURITY_PROVIDER_KEY: "db",
+                API_SECURITY_REFRESH_KEY: refresh,
+            },
         )
 
     def login(self, client, username, password):
-        # Login with default admin
         rv = self._login(client, username, password)
         try:
             return json.loads(rv.data.decode("utf-8")).get("access_token")
         except Exception:
             return rv
 
-    def browser_login(self, client, username, password):
-        # Login with default admin
+    def browser_login(
+        self,
+        client: FlaskClient,
+        username: str,
+        password: str,
+        next_url: Optional[str] = None,
+        follow_redirects: bool = True,
+    ) -> Response:
+        login_url = "/login/"
+        if next_url:
+            login_url = f"{login_url}?next={next_url}"
         return client.post(
-            "/login/",
+            login_url,
             data=dict(username=username, password=password),
-            follow_redirects=True,
+            follow_redirects=follow_redirects,
         )
 
+    def assert_response(
+        self,
+        response: List[Dict[str, Any]],
+        expected_results: List[Dict[str, Any]],
+        exclude_cols: Optional[List[str]] = None,
+    ):
+        exclude_cols = exclude_cols or []
+        for idx, expected_result in enumerate(expected_results):
+            for field_name, field_value in expected_result.items():
+                if field_name not in exclude_cols:
+                    self.assertEqual(
+                        response[idx][field_name], expected_result[field_name]
+                    )
+
     @staticmethod
     def browser_logout(client):
         return client.get("/logout/")
